# Copyright (C) 2012 Anaconda, Inc
# SPDX-License-Identifier: BSD-3-Clause
"""Base exceptions."""
from ..deprecations import deprecated

deprecated.module("24.3", "24.9", addendum="Nothing to import.")
